@@ -74,10 +74,10 @@ typedef enum
 
 /* Returns error string or NULL if no error */
 // FIXME: Global variable
-const char *mp_get_error();
+const char *mp_get_error(void);
 
 /* Clear any current error */
-void mp_clear_error();
+void mp_clear_error(void);
 
 /* Returns:
  *  0 if x == y
@@ -276,8 +276,9 @@ uint64_t mp_cast_to_unsigned_int(const MPNumber *x);
  * The numbers are written in 'base' (e.g. 10).
  * If 'trim_zeroes' is non-zero then strip off trailing zeroes.
  * Fractional components are truncated at 'max_digits' digits.
+ * The radix character (e.g. '.' or ',') can be specified.
  */
-void   mp_cast_to_string(const MPNumber *x, int default_base, int base, int max_digits, bool trim_zeroes, char *buffer, int buffer_length);
+void   mp_cast_to_string(const MPNumber *x, int default_base, int base, int max_digits, bool trim_zeroes, gunichar radix, char *buffer, int buffer_length);
 
 /* Converts x to a string representation in exponential form.
  * The string is written into 'buffer' which is guaranteed to be at least 'buffer_length' octets in size.
@@ -286,7 +287,7 @@ void   mp_cast_to_string(const MPNumber *x, int default_base, int base, int max_
  * If 'trim_zeroes' is non-zero then strip off trailing zeroes.
  * Fractional components are truncated at 'max_digits' digits.
  */
-void   mp_cast_to_exponential_string(const MPNumber *x, int default_base, int base, int max_digits, bool trim_zeroes, bool eng_format, char *buffer, int buffer_length);
+void   mp_cast_to_exponential_string(const MPNumber *x, int default_base, int base, int max_digits, bool trim_zeroes, gunichar radix, bool eng_format, char *buffer, int buffer_length);
 
 /* Sets z = sin x */
 void   mp_sin(const MPNumber *x, MPAngleUnit unit, MPNumber *z);
